// this :
#include <BatchLab/Core/BaseFitter.h>

#include <AIDA/ITree.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IProfile1D.h>
#include <AIDA/IAxis.h>
#include <AIDA/IFunctionFactory.h>
#include <AIDA/IFunction.h>
#include <AIDA/IDataPointSet.h>
#include <AIDA/IDataPoint.h>
#include <AIDA/IMeasurement.h>
#include <AIDA/IAnnotation.h>
#include <AIDA/ICloud1D.h>
#include <AIDA/ICloud2D.h>
#include <AIDA/ITuple.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/IEvaluator.h>
#include <AIDA/IFilter.h>

#include <Slash/Core/IWriter.h>
#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>
#include <Lib/Debug.h>

#include <BatchLab/Core/FitResult.h>
#include <BatchLab/Core/FitData.h>

#include <BatchLab/Core/Histogram.h> //FIXME

#include <math.h>

#define MAXIMUM(a,b) ((a)>(b)?a:b)
#define NotFound (-1)

//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseFitter::setFitInfos(
 AIDA::IAnnotation& aAnnotation
,AIDA::IFitResult& aResult
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aAnnotation.removeItem("fit.quality");
  aAnnotation.addItem("fit.quality",Lib::smanip::tostring(aResult.quality()));

  aAnnotation.removeItem("fit.ndf");
  aAnnotation.addItem("fit.ndf",Lib::smanip::tostring(aResult.ndf()));

  std::vector<std::string> names = aResult.fittedParameterNames();
  std::vector<double> values = aResult.fittedParameters();
  std::vector<double> errs = aResult.errors();
  if(names.size()!=values.size()) return;
  if(names.size()!=errs.size()) return;

  unsigned int namen = names.size();
  for(unsigned index=0;index<namen;index++){
    std::string key = "fit.param."+names[index];
    aAnnotation.removeItem(key);
    aAnnotation.addItem(key,
      Lib::smanip::tostring(values[index])+" "+
      Lib::smanip::tostring(errs[index]));
  }
}

//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseFitter::BaseFitter(
 Slash::Core::IWriter& aPrinter
,AIDA::IFunctionFactory& aFunctionFactory
,AIDA::ITupleFactory& aTupleFactory
,const std::string& aEngine
)
:fPrinter(aPrinter)
,fFunctionFactory(aFunctionFactory)
,fTupleFactory(aTupleFactory)
,fEngineName(aEngine)
,fCounter(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::BaseFitter");
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseFitter::~BaseFitter(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::decrement("BatchLab::BaseFitter");
}
//////////////////////////////////////////////////////////////////////////////
void* BatchLab::BaseFitter::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(AIDA::IFitter)
  else if_Lib_SCast(BatchLab::BaseFitter)
  else return 0;
}

//////////////////////////////////////////////////////////////////////////////
template <class T> static AIDA::IFitResult* BatchLab_BaseFitter_fit1D(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,T& aHistogram //histo1D or profile1D
,AIDA::IFunction& aFunction
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aHistogram.entries()==0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " histogram has no entries." << Lib::endl;
    return 0;
  }

  int nbin = aHistogram.axis().bins();
  if(nbin<=0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " has no bins." << Lib::endl;
    return 0;
  }

  int number = 0;
 {for(int count=0;count<nbin;count++) { 
    int nent = aHistogram.binEntries(count);
    if(nent==0) continue;  // Do not take bin without entry.
    number++;
  }}

  std::vector<double> pt;
  pt.resize(1,0);
  std::vector< std::vector<double> > pts;
  pts.resize(number,pt);

  std::vector<double> vs(number);
  std::vector<double> es(number);
  int bini = 0;
  for(int index=0;index<nbin;index++) { 
    int nent = aHistogram.binEntries(index);
    if(nent==0) continue;  // Do not take bin without entry.
    // Take middle of bin.
    double xx = 
      (aHistogram.axis().binLowerEdge(index)+
       aHistogram.axis().binUpperEdge(index))/2.;
    double val = aHistogram.binHeight(index);
    double err = ::sqrt(::fabs(val)); // Take statistical error.
    if(err==0.) { 
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " null error for bin " << index 
          << " x " << xx << " value " << val
          << ". Set error to 0.01."
          << Lib::endl;
      //return 0;
      err = 0.01; //FIXME
    }
    pts[bini][0] = xx;
    vs[bini] = val;
    es[bini] = err;
    bini++;
  }
  if(pts.size()==0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " no starting points." << Lib::endl;
    return 0;
  }

  double chi2;
  std::vector<double> errs;
  if(!aFitter.baseFit(aFunction,pts,vs,es,chi2,errs)) return 0;

  const std::vector<double>& func_params = aFunction.parameters();
  int degreeOfFreedom = pts.size()-func_params.size();
  double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

  BatchLab::FitResult* result = new BatchLab::FitResult
    (aFunction,errs,degreeOfFreedom,quality);

  BatchLab::BaseFitter::setFitInfos(aHistogram.annotation(),*result);

  return result;
}

//////////////////////////////////////////////////////////////////////////////
template <class T> static AIDA::IFitResult* BatchLab_BaseFitter_fit2D(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,T& aHistogram //histo2D or profile2D
,AIDA::IFunction& aFunction
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aHistogram.entries()==0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " histogram has no entries." << Lib::endl;
    return 0;
  }

  int nx = aHistogram.xAxis().bins();
  if(nx<=0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " has no x bins." << Lib::endl;
    return 0;
  }

  int ny = aHistogram.yAxis().bins();
  if(ny<=0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " has no y bins." << Lib::endl;
    return 0;
  }

  int number = 0;
 {for(int i=0;i<nx;i++) { 
    for(int j=0;j<ny;j++) { 
      int nent = aHistogram.binEntries(i,j);
      if(nent==0) continue;  // Do not take bin without entry.
      number++;
    }
  }}

  std::vector<double> pt;
  pt.resize(2,0);
  std::vector< std::vector<double> > pts;
  pts.resize(number,pt);

  std::vector<double> vs(number);
  std::vector<double> es(number);
  int bini = 0;
 {for(int i=0;i<nx;i++) { 
    // Take middle of bin.
    double xx = 
      (aHistogram.xAxis().binLowerEdge(i)+
       aHistogram.xAxis().binUpperEdge(i))/2.;
    for(int j=0;j<ny;j++) { 
      int nent = aHistogram.binEntries(i,j);
      if(nent==0) continue;  // Do not take bin without entry.

      double yy = 
        (aHistogram.yAxis().binLowerEdge(j)+
         aHistogram.yAxis().binUpperEdge(j))/2.;

      double val = aHistogram.binHeight(i,j);
      double err = ::sqrt(::fabs(val)); // Take statistical error.
      if(err==0.) { 
        Lib::Out out(aPrinter);
        out << "BatchLab::BaseFitter::fit :"
            << " null error for bin"
            << " i " << i << " x " << xx 
            << " j " << j << " y " << yy
            << " value " << val
            << ". Set error to 0.01."
            << Lib::endl;
        //return 0;
        err = 0.01; //FIXME
      }
      pts[bini][0] = xx;
      pts[bini][1] = yy;
      vs[bini] = val;
      es[bini] = err;
      bini++;
    }
  }}
  if(pts.size()==0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " no starting points." << Lib::endl;
    return 0;
  }

  double chi2;
  std::vector<double> errs;
  if(!aFitter.baseFit(aFunction,pts,vs,es,chi2,errs)) return 0;

  const std::vector<double>& func_params = aFunction.parameters();
  int degreeOfFreedom = pts.size()-func_params.size();
  double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

  BatchLab::FitResult* result = new BatchLab::FitResult
    (aFunction,errs,degreeOfFreedom,quality);

  BatchLab::BaseFitter::setFitInfos(aHistogram.annotation(),*result);

  return result;
}

//////////////////////////////////////////////////////////////////////////////
static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::ICloud1D& aCloud
,AIDA::IFunction& aFunction
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aCloud.isConverted()) {

    //break const :
    AIDA::IHistogram1D* histo = (AIDA::IHistogram1D*)&(aCloud.histogram());
    return BatchLab_BaseFitter_fit1D(aFitter,aPrinter,*histo,aFunction);
    
  } else {

    int number = aCloud.entries();
    if(number==0) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit(Cloud1D) :"
          << " cloud has no entries." << Lib::endl;
      return 0;
    }

    std::vector<double> pt;
    pt.resize(1,0);
    std::vector< std::vector<double> > pts;
    pts.resize(number,pt);

    std::vector<double> vs(number);
    std::vector<double> es(number);

    for(unsigned int index=0;index<number;index++) { 
      double xx = aCloud.value(index);
      double val = aCloud.weight(index);
      double err = ::sqrt(::fabs(val)); // Take statistical error.
      if(err==0.) { 
        Lib::Out out(aPrinter);
        out << "BatchLab::BaseFitter::fit(Cloud1D) :"
            << " null error for point " << index
            << " x " << xx << " value " << val
            << ". Set error to 0.01."
            << Lib::endl;
        //return 0;
        err = 0.01; //FIXME
      }
      pts[index][0] = xx;
      vs[index] = val;
      es[index] = err;
    }

    double chi2;
    std::vector<double> errs;
    if(!aFitter.baseFit(aFunction,pts,vs,es,chi2,errs)) return 0;

    const std::vector<double>& func_params = aFunction.parameters();
    int degreeOfFreedom = pts.size()-func_params.size();
    double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

    BatchLab::FitResult* result = new BatchLab::FitResult
      (aFunction,errs,degreeOfFreedom,quality);

    BatchLab::BaseFitter::setFitInfos(aCloud.annotation(),*result);

    return result;
  }
}
//////////////////////////////////////////////////////////////////////////////
static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::ICloud2D& aCloud
,AIDA::IFunction& aFunction
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aCloud.isConverted()) {

    //return BatchLab_BaseFitter_fit
    //         (aFitter,aPrinter,aCloud.histogram(),aFunction);
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit(Cloud2D) :"
        << " don't know how to fit an histogram 2D." << Lib::endl;
    return 0;

  } else {

    int number = aCloud.entries();
    if(number==0) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit(Cloud2D) :"
          << " cloud has no entries." << Lib::endl;
      return 0;
    }

    std::vector<double> pt;
    pt.resize(1,0);
    std::vector< std::vector<double> > pts;
    pts.resize(number,pt);

    std::vector<double> vs(number);
    std::vector<double> es(number);

    for(unsigned int index=0;index<number;index++) { 
      double xx = aCloud.valueX(index);
      double val = aCloud.valueY(index);
      double err = ::sqrt(::fabs(val)); // Take statistical error.
      if(err==0.) { 
        Lib::Out out(aPrinter);
        out << "BatchLab::BaseFitter::fit(Cloud2D) :"
            << " null error for point " << index
            << " x " << xx << " value " << val
            << ". Set error to 0.01."
            << Lib::endl;
        //return 0;
        err = 0.01; //FIXME
      }
      pts[index][0] = xx;
      vs[index] = val;
      es[index] = err;
    }

    double chi2;
    std::vector<double> errs;
    if(!aFitter.baseFit(aFunction,pts,vs,es,chi2,errs)) return 0;

    const std::vector<double>& func_params = aFunction.parameters();
    int degreeOfFreedom = pts.size()-func_params.size();
    double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

    BatchLab::FitResult* result = new BatchLab::FitResult
      (aFunction,errs,degreeOfFreedom,quality);

    BatchLab::BaseFitter::setFitInfos(aCloud.annotation(),*result);

    return result;
  }
}

//////////////////////////////////////////////////////////////////////////////
static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::ITuple& aTuple
,AIDA::IEvaluator& aEvaluator
,AIDA::IFilter& aFilter
,AIDA::IFunction& aFunction
){
  //
  // for one column we fit {(row, col[row])}
  //
  int number = aTuple.rows();
  if(number<=0) {
    Lib::Out out(aPrinter);
    out << "BatchLab_BaseFitter_fit :"
        << " Tuple " 
        << Lib::sout(aTuple.title()) << " empty." 
        << Lib::endl;
    return 0;
  }

  if(!aEvaluator.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab_BaseFitter_fit :"
        << " Evaluator initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return 0;
  }

  if(!aFilter.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab_BaseFitter_fit :"
        << " Filter initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return 0;
  }

  std::vector<double> pt;
  pt.resize(1,0);
  std::vector< std::vector<double> > pts;
  pts.resize(number,pt);

  std::vector<double> vs(number);
  std::vector<double> es(number);

  unsigned int index = 0;
  aTuple.start();
  while(aTuple.next()) {
    if(!aFilter.accept()) continue;
    double xx = double(index);
    double val = aEvaluator.evaluateDouble();
    double err = ::sqrt(::fabs(val)); // Take statistical error.
    if(err==0.) { 
      Lib::Out out(aPrinter);
      out << "BatchLab_BaseFitter_fit :"
          << " null error for point " << index
          << " x " << xx << " value " << val
          << ". Set error to 0.01."
          << Lib::endl;
      //return 0;
      err = 0.01; //FIXME
    }
    pts[index][0] = xx;
    vs[index] = val;
    es[index] = err;
    index++;
  }
  pts.resize(index);
  vs.resize(index);
  es.resize(index);

  double chi2;
  std::vector<double> errs;
  if(!aFitter.baseFit(aFunction,pts,vs,es,chi2,errs)) return 0;

  const std::vector<double>& func_params = aFunction.parameters();
  int degreeOfFreedom = pts.size()-func_params.size();
  double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

  BatchLab::FitResult* result = new BatchLab::FitResult
    (aFunction,errs,degreeOfFreedom,quality);

  BatchLab::BaseFitter::setFitInfos(aTuple.annotation(),*result);

  return result;
}

//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::IDataPointSet& aDataPointSet
,int a_x_index
,int a_v_index
,AIDA::IFunction& aFunction
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int number = aDataPointSet.size();
  if(number<=0) return 0;

  int dim = aDataPointSet.dimension();

  if((a_x_index<0)||(dim<=a_x_index)) {
    Lib::Out out(aPrinter);
    out << "BatchLab_BaseFitter_fit :"
        << " dataPointSet index " << a_x_index
        << " is out of range. Dimension is " << dim << "."
        << Lib::endl;
    return 0;
  }
  if((a_v_index<0)||(dim<=a_v_index)) {
    Lib::Out out(aPrinter);
    out << "BatchLab_BaseFitter_fit :"
        << " dataPointSet index " << a_v_index
        << " is out of range. Dimension is " << dim << "."
        << Lib::endl;
    return 0;
  }

  std::vector<double> pt;
  pt.resize(1,0);
  std::vector< std::vector<double> > pts;
  pts.resize(number,pt);

  std::vector<double> vs(number);
  std::vector<double> es(number);
  for(int index=0;index<number;index++) {
    double x = aDataPointSet.point(index)->coordinate(a_x_index)->value();
    double v = aDataPointSet.point(index)->coordinate(a_v_index)->value();
    double e = ::sqrt(::fabs(v));
    if(e==0) e = 0.01; //WARNING : must not be 0.
    pts[index][0] = x;
    vs[index] = v;
    es[index] = e;
  }
  
  double chi2;
  std::vector<double> errs;
  if(!aFitter.baseFit(aFunction,pts,vs,es,chi2,errs)) return 0;

  const std::vector<double>& func_params = aFunction.parameters();
  int degreeOfFreedom = pts.size()-func_params.size();
  double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

  BatchLab::FitResult* result = 
    new BatchLab::FitResult(aFunction,errs,degreeOfFreedom,quality);

  BatchLab::BaseFitter::setFitInfos(aDataPointSet.annotation(),*result);

  return result;
}

//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IBaseHistogram& aHistogram
,AIDA::IFunction& aFunction
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(AIDA::IHistogram1D* histo = Lib_Cast(aHistogram,AIDA::IHistogram1D)) {
    return BatchLab_BaseFitter_fit1D(*this,fPrinter,*histo,aFunction);
  } else if(AIDA::IProfile1D* prof = Lib_Cast(aHistogram,AIDA::IProfile1D)) {
    return BatchLab_BaseFitter_fit1D(*this,fPrinter,*prof,aFunction);
  } else if(AIDA::ICloud1D* cloud = Lib_Cast(aHistogram,AIDA::ICloud1D)) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,*cloud,aFunction);
  } else if(AIDA::ICloud2D* cloud = Lib_Cast(aHistogram,AIDA::ICloud2D)) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,*cloud,aFunction);
  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " base histogram " << Lib::sout(aHistogram.title())
        << " is not an AIDA::IHistogram1D, AIDA::IProfile1D,"
        << " AIDA::ICloud1D, an AIDA::ICloud2D." 
        << Lib::endl;
    return 0;
  }
}

//////////////////////////////////////////////////////////////////////////////
template <class T> static bool HalfLowerMeanPoint (
 T& aHisto
,double& aX
,double& aV
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int binn = aHisto.axis().bins();
  int end = binn/2;
  unsigned int n = 0;
  aX = 0;
  aV = 0;
  for(int count=0;count<end;count++) { 
    if(aHisto.binEntries(count)==0) continue;  // Do not take empty bins.
    // Take middle of Bin.
    aX += 
      (aHisto.axis().binLowerEdge(count)+aHisto.axis().binUpperEdge(count))/2.;
    aV += aHisto.binHeight(count);
    n++;
  }
  if(!n) {
    aX = 0;
    aV = 0;
    return false;
  }
  aX /= n;
  aV /= n;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
template <class T> static bool HalfUpperMeanPoint(
 T& aHisto
,double& aX
,double& aV
)
//////////////////////////////////////////////////////////////////////////////
//  Used to find an end point for linear, exponential fit.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int binn = aHisto.axis().bins();
  int begin = binn/2;
  unsigned int n = 0;
  aX = 0;
  aV = 0;
  for(int count=begin;count<binn;count++) {
    if(aHisto.binEntries(count)==0) continue;  // Do not take empty bins.
    // Take middle of Bin.
    aX += 
      (aHisto.axis().binUpperEdge(count)+aHisto.axis().binLowerEdge(count))/2.;
    aV += aHisto.binHeight(count);
    n++;
  }
  if(!n) {
    aX = 0;
    aV = 0;
    return false;
  }
  aX /= n;
  aV /= n;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
static bool HalfLowerMeanPoint (
 AIDA::ICloud1D& aCloud
,double& aX
,double& aV
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double m = aCloud.mean();
  unsigned int n = 0;
  aX = 0;
  aV = 0;
 {int number = aCloud.entries();
  for(unsigned int index=0;index<number;index++) { 
    double xx = aCloud.value(index);
    if(xx>m) continue;
    aX += xx;
    aV += aCloud.weight(index);
    n++;
  }}
  if(!n) {
    aX = 0;
    aV = 0;
    return false;
  }
  aX /= n;
  aV /= n;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
static bool HalfUpperMeanPoint (
 AIDA::ICloud1D& aCloud
,double& aX
,double& aV
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double m = aCloud.mean();
  unsigned int n = 0;
  aX = 0;
  aV = 0;
 {int number = aCloud.entries();
  for(unsigned int index=0;index<number;index++) { 
    double xx = aCloud.value(index);
    if(xx<=m) continue;
    aX += xx;
    aV += aCloud.weight(index);
    n++;
  }}
  if(!n) {
    aX = 0;
    aV = 0;
    return false;
  }
  aX /= n;
  aV /= n;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
template <class T> static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::IFunctionFactory& aFunctionFactory
,int& aCounter
,T& aHistogram
,const std::string& aModel
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //AIDA::IAnnotation& annotation = aHistogram.annotation();
  //std::string name = annotation.value("fit_function_name");

  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitHistogram_%d",aCounter);
  aCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    double h = aHistogram.maxBinHeight() - aHistogram.minBinHeight();
    double m = aHistogram.mean();
    double w = aHistogram.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    //NOTE : fit.function is used in OpenPAW /HISTOGRAM/PLOT.
    aHistogram.annotation().removeItem("fit.function");
    aHistogram.annotation().addItem("fit.function",name);

    return BatchLab_BaseFitter_fit1D(aFitter,aPrinter,aHistogram,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    double h = aHistogram.maxBinHeight() - aHistogram.minBinHeight();
    double m = aHistogram.mean();
    double w = aHistogram.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    aHistogram.annotation().removeItem("fit.function");
    aHistogram.annotation().addItem("fit.function",name);

    return BatchLab_BaseFitter_fit1D(aFitter,aPrinter,aHistogram,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    double xx1,xx2;
    double val1,val2;
    bool status1 = HalfLowerMeanPoint(aHistogram,xx1,val1);
    bool status2 = HalfUpperMeanPoint(aHistogram,xx2,val2);
    if( !status1 || !status2 || ((xx2-xx1)==0.) ) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    if(!val1) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    double dx = xx2 - xx1;
    if(!dx) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    // Deduce p0,p1 of p0*exp(p1*x) with the two upper points :
    double p1 = ::log(val2/val1) / dx;
    double p0 = ::exp(p1*xx1);
    if(!p0) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }    
    p0 = val1 / p0;

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = p0;
    ps[1] = p1;

    function->setParameters(ps);

    aHistogram.annotation().removeItem("fit.function");
    aHistogram.annotation().addItem("fit.function",name);

    return BatchLab_BaseFitter_fit1D(aFitter,aPrinter,aHistogram,*function);

  } else if(model=="EHBOOK") {

    double xx1,xx2;
    double val1,val2;
    bool status1 = HalfLowerMeanPoint(aHistogram,xx1,val1);
    bool status2 = HalfUpperMeanPoint(aHistogram,xx2,val2);
    if( !status1 || !status2 || ((xx2-xx1)==0.) ) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    
    // Deduce a,b of exp(a+bx) with the two upper points :
    double b = ::log(val2/val1) / ( xx2 - xx1);
    double a = ::log( val1 / exp ( b * xx1 ) );

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"EHBOOK");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = a;
    ps[1] = b;
    function->setParameters(ps);

    aHistogram.annotation().removeItem("fit.function");
    aHistogram.annotation().addItem("fit.function",name);

    return BatchLab_BaseFitter_fit1D(aFitter,aPrinter,aHistogram,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    double h = aHistogram.maxBinHeight() - aHistogram.minBinHeight();
    double m = aHistogram.mean();
    double w = aHistogram.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    std::vector<double> ps;
    ps.resize(degree+1,1);
    if(degree==2) {
      ps[0] = h;
      ps[1] = m;
      ps[2] = w;
    } 
    function->setParameters(ps);

    aHistogram.annotation().removeItem("fit.function");
    aHistogram.annotation().addItem("fit.function",name);

    return BatchLab_BaseFitter_fit1D(aFitter,aPrinter,aHistogram,*function);

  } 

  Lib::Out out(aPrinter);
  out << "BatchLab::BaseFitter::fit :"
      << " unknown fit model " << Lib::sout(model)
      << Lib::endl;

  return 0;
}

//////////////////////////////////////////////////////////////////////////////
static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::IFunctionFactory& aFunctionFactory
,int& aCounter
,AIDA::ICloud1D& aCloud
,const std::string& aModel
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitCloud_%d",aCounter);
  aCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    double h = 0;
   {int number = aCloud.entries();
    for(unsigned int index=0;index<number;index++) { 
      h = MAXIMUM(h,aCloud.weight(index));
    }}
    double m = aCloud.mean();
    double w = aCloud.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aCloud,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    double h = 0;
   {int number = aCloud.entries();
    for(unsigned int index=0;index<number;index++) { 
      h = MAXIMUM(h,aCloud.weight(index));
    }}
    double m = aCloud.mean();
    double w = aCloud.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aCloud,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    double xx1,xx2;
    double val1,val2;
    bool status1 = HalfLowerMeanPoint(aCloud,xx1,val1);
    bool status2 = HalfUpperMeanPoint(aCloud,xx2,val2);
    if( !status1 || !status2 || ((xx2-xx1)==0.) ) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    
    // Deduce p0,p1 of p0*exp(p1*x) with the two upper points :
    double p1 = ::log(val2/val1) / ( xx2 - xx1);
    double p0 = val1 / exp ( p1 * xx1 );

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = p0;
    ps[1] = p1;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aCloud,*function);

  } else if(model=="EHBOOK") {

    double xx1,xx2;
    double val1,val2;
    bool status1 = HalfLowerMeanPoint(aCloud,xx1,val1);
    bool status2 = HalfUpperMeanPoint(aCloud,xx2,val2);
    if( !status1 || !status2 || ((xx2-xx1)==0.) ) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    
    // Deduce a,b of exp(a+bx) with the two upper points :
    double b = ::log(val2/val1) / ( xx2 - xx1);
    double a = ::log( val1 / exp ( b * xx1 ) );

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"EHBOOK");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = a;
    ps[1] = b;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aCloud,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    double h = 0;
   {int number = aCloud.entries();
    for(unsigned int index=0;index<number;index++) { 
      h = MAXIMUM(h,aCloud.weight(index));
    }}
    double m = aCloud.mean();
    double w = aCloud.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    std::vector<double> ps;
    ps.resize(degree+1,1);
    if(degree==2) {
      ps[0] = h;
      ps[1] = m;
      ps[2] = w;
    } 
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aCloud,*function);

  } 

  Lib::Out out(aPrinter);
  out << "BatchLab::BaseFitter::fit :"
      << " unknown fit model " << Lib::sout(model)
      << Lib::endl;

  return 0;
}

//////////////////////////////////////////////////////////////////////////////
static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::IFunctionFactory& aFunctionFactory
,int& aCounter
,AIDA::ITuple& aTuple
,AIDA::IEvaluator& aEvaluator
,AIDA::IFilter& aFilter
,const std::string& aModel
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitTuple_%d",aCounter);
  aCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    double h = 0;
    double mean = 0;
    double rms = 0;
   {double Sw = 0;
    double Sxw = 0;
    double Sx2w = 0;
    unsigned int index = 0;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      double x = double(index);
      double w = aEvaluator.evaluateDouble();
      h = MAXIMUM(h,w);
      Sw += w;
      double xw = x * w;
      Sxw += xw;
      Sx2w += x * xw;
      index++;
    }
    if(Sw!=0) {
      mean = Sxw/Sw;
      rms = ::sqrt(::fabs( (Sx2w / Sw) - mean * mean));
    }}

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = mean;
    ps[2] = rms;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,
      aTuple,aEvaluator,aFilter,*function);

/*uuu
  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    double h = 0;
   {int number = aCloud.entries();
    for(unsigned int index=0;index<number;index++) { 
      h = MAXIMUM(h,aCloud.weight(index));
    }}
    double m = aCloud.mean();
    double w = aCloud.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,
      aTuple,aEvaluator,aFilter,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    double xx1,xx2;
    double val1,val2;
    bool status1 = HalfLowerMeanPoint(aCloud,xx1,val1);
    bool status2 = HalfUpperMeanPoint(aCloud,xx2,val2);
    if( !status1 || !status2 || ((xx2-xx1)==0.) ) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    
    // Deduce p0,p1 of p0*exp(p1*x) with the two upper points :
    double p1 = ::log(val2/val1) / ( xx2 - xx1);
    double p0 = val1 / exp ( p1 * xx1 );

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = p0;
    ps[1] = p1;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,
      aTuple,aEvaluator,aFilter,*function);

  } else if(model=="EHBOOK") {

    double xx1,xx2;
    double val1,val2;
    bool status1 = HalfLowerMeanPoint(aCloud,xx1,val1);
    bool status2 = HalfUpperMeanPoint(aCloud,xx2,val2);
    if( !status1 || !status2 || ((xx2-xx1)==0.) ) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    
    // Deduce a,b of exp(a+bx) with the two upper points :
    double b = ::log(val2/val1) / ( xx2 - xx1);
    double a = ::log( val1 / exp ( b * xx1 ) );

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"EHBOOK");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = a;
    ps[1] = b;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,
      aTuple,aEvaluator,aFilter,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    double h = 0;
   {int number = aCloud.entries();
    for(unsigned int index=0;index<number;index++) { 
      h = MAXIMUM(h,aCloud.weight(index));
    }}
    double m = aCloud.mean();
    double w = aCloud.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    std::vector<double> ps;
    ps.resize(degree+1,1);
    if(degree==2) {
      ps[0] = h;
      ps[1] = m;
      ps[2] = w;
    } 
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,
      aTuple,aEvaluator,aFilter,*function);

*/
  } 

  Lib::Out out(aPrinter);
  out << "BatchLab::BaseFitter::fit :"
      << " unknown fit model " << Lib::sout(model)
      << Lib::endl;

  return 0;
}


//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IBaseHistogram& aHistogram
,const std::string& aModel
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(AIDA::IHistogram1D* histo = Lib_Cast(aHistogram,AIDA::IHistogram1D)) {
    return BatchLab_BaseFitter_fit
             (*this,fPrinter,fFunctionFactory,fCounter,*histo,aModel);
  } else if(AIDA::IProfile1D* prof = Lib_Cast(aHistogram,AIDA::IProfile1D)) {
    return BatchLab_BaseFitter_fit
             (*this,fPrinter,fFunctionFactory,fCounter,*prof,aModel);

  } else if(AIDA::ICloud1D* cloud = Lib_Cast(aHistogram,AIDA::ICloud1D)) {
    return BatchLab_BaseFitter_fit
             (*this,fPrinter,fFunctionFactory,fCounter,*cloud,aModel);
//} else if(AIDA::ICloud2D* cloud = Lib_Cast(aHistogram,AIDA::ICloud2D)) {
  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit(model) :"
        << " base histogram " << Lib::sout(aHistogram.title())
        << " is not an AIDA::IHistogram1D, AIDA::IProfile1D,"
        << " AIDA::ICloud1D." 
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IDataPointSet& aDataPointSet
,AIDA::IFunction& aFunction
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return BatchLab_BaseFitter_fit(*this,fPrinter,aDataPointSet,0,1,aFunction);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IDataPointSet& aDataPointSet
,const std::string& aModel
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitDataPointSet_%d",fCounter);
  fCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    double h = 1; //FIXME
    double m = 0;
    double w = 0;

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return fit(aDataPointSet,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    double h = 1; //FIXME
    double m = 0;
    double w = 0;

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return fit(aDataPointSet,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    double p0 = 1;
    double p1 = 1;

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = p0;
    ps[1] = p1;
    function->setParameters(ps);

    return fit(aDataPointSet,*function);

  } else if(model=="EHBOOK") {

    double a = 0;
    double b = 1;

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"EHBOOK");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = a;
    ps[1] = b;
    function->setParameters(ps);

    return fit(aDataPointSet,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    std::vector<double> ps;
    ps.resize(degree+1,1);
    function->setParameters(ps);
    return fit(aDataPointSet,*function);

  } 

  Lib::Out out(fPrinter);
  out << "BatchLab::BaseFitter::fit : unknown fit model " << Lib::sout(model)
      << Lib::endl;

  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IBaseHistogram& aHistogram
,const std::string& aModel
,const std::vector<double>& aParams
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  AIDA::IHistogram1D* histo = Lib_Cast(aHistogram,AIDA::IHistogram1D);
  if(!histo) {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " histogram " << Lib::sout(aHistogram.title())
        << " is not an AIDA::IHistogram1D." 
        << Lib::endl;
    return 0;
  }

  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitHistogram_%d",fCounter);
  fCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    if(aParams.size()!=3) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (3 expected)."
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    if(aParams.size()!=3) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (3 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    if(aParams.size()!=2) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (2 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else if(model=="EHBOOK") {

    if(aParams.size()!=2) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (2 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"EBOOK");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    if((int)aParams.size()!=(degree+1)) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (" << (degree+1) << " expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " unknown fit model " << Lib::sout(model)
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IDataPointSet& aDataPointSet
,const std::string& aModel
,const std::vector<double>& aParams
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitDataPointSet_%d",fCounter);
  fCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    if(aParams.size()!=3) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (3 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aDataPointSet,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    if(aParams.size()!=3) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (3 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aDataPointSet,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    if(aParams.size()!=2) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (2 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aDataPointSet,*function);

  } else if(model=="EHBOOK") {

    if(aParams.size()!=2) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (2 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"EHBOOK");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aDataPointSet,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    if((int)aParams.size()!=(degree+1)) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (" << (degree+1) << " expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    function->setParameters(aParams);
    return fit(aDataPointSet,*function);

  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit : unknown fit model " << Lib::sout(model)
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitParameterSettings& BatchLab::BaseFitter::fitParameterSettings(
 const std::string& aName
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fFitParameterSettings.size();
  for(unsigned int index=0;index<number;index++) {
    if(fFitParameterSettings[index].name()==aName) {
      return fFitParameterSettings[index];
    }
  }
  // Not found, create one :
  fFitParameterSettings.push_back(FitParameterSettings(aName));
  return fFitParameterSettings[fFitParameterSettings.size()-1];
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> BatchLab::BaseFitter::listParameterSettings(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fFitParameterSettings.size();
  std::vector<std::string> names(number);
  for(unsigned int index=0;index<number;index++) {
    names[index] = fFitParameterSettings[index].name();
  }
  return names;
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseFitter::resetParameterSettings(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fParameterSettings.clear(); //FIXME : is it what AIDA semanticly wants ?
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseFitter::engineName(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fEngineName;
}

bool BatchLab::BaseFitter::setEngine(
 const std::string& /*name*/
) 
{
  return false;
}
bool BatchLab::BaseFitter::setConstraint(
 const std::string& /*expr*/
) 
{
  return false;
}
std::vector<std::string> BatchLab::BaseFitter::constraints(
) 
{
  return fConstraints;
}
void BatchLab::BaseFitter::resetConstraints(
) 
{
}
bool BatchLab::BaseFitter::setFitMethod(
 const std::string& /*name*/
) 
{
  return false;
}
std::string BatchLab::BaseFitter::fitMethodName(
) 
{
  return fFitMethodName;
}
AIDA::IDataPointSet* BatchLab::BaseFitter::createScan1D(
 AIDA::IFitData& /*d*/
,AIDA::IFunction& /*f*/
,const std::string& /*par*/
,int /*npts*/
,double /*pmin*/
,double /*pmax*/
) 
{
  return 0;
}
AIDA::IDataPointSet* BatchLab::BaseFitter::createContour(
 AIDA::IFitData& /*d*/
,AIDA::IFitResult& /*r*/
,const std::string& /*par1*/
,const std::string& /*par2*/
,int /*npts*/
,double /*up*/
) 
{
  return 0;
}
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
void BatchLab::BaseFitter::setUseFunctionGradient(bool /*useGrad*/) 
{
}
#else
bool BatchLab::BaseFitter::setUseFunctionGradient(bool /*useGrad*/) 
{
  return false;
}
#endif
bool BatchLab::BaseFitter::useFunctionGradient(
) 
{
  return false;
}

//FIXME
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
AIDA::IFitResult* BatchLab::BaseFitter::fit(AIDA::IFitData& aFitData,AIDA::IFunction& aFunction){
  //FIXME : there is no IFitData::cast !
  //BatchLab::FitData* fitData = Lib_Cast(aFitData,BatchLab::FitData);
  BatchLab::FitData* fitData = dynamic_cast<BatchLab::FitData*>(&aFitData);
  if(!fitData) {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :" 
        << " passed AIDA::IFitData is not a BatchLab::FitData."
        << Lib::endl;
    return 0;
  }

 {const AIDA::IHistogram1D* histo = fitData->histogram1D();
  if(histo) {
    return BatchLab_BaseFitter_fit1D(*this,fPrinter,
              *const_cast<AIDA::IHistogram1D*>(histo),aFunction);
  }}

 {const AIDA::IHistogram2D* histo = fitData->histogram2D();
  if(histo) {
    return BatchLab_BaseFitter_fit2D(*this,fPrinter,
              *const_cast<AIDA::IHistogram2D*>(histo),aFunction);
  }}

 {int x,v;
  const AIDA::IDataPointSet* dps = fitData->dps(x,v);
  if(dps) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,
              *const_cast<AIDA::IDataPointSet*>(dps),x,v,aFunction);
  }}

 {std::vector<std::string> cols;
  const AIDA::ITuple* tuple = fitData->tuple(cols);
  if(tuple) {
    if(cols.size()!=1) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :" 
          << " can't handle today more than one column."
          << Lib::endl;
      return 0;
    }
    AIDA::IEvaluator* evaluator = fTupleFactory.createEvaluator(cols[0]);
    if(!evaluator) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :" 
          << " can't create evaluator."
          << Lib::endl;
      return 0;
    }
    AIDA::IFilter* filter = fTupleFactory.createFilter("");
    if(!filter) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :" 
          << " can't create filter."
          << Lib::endl;
      return 0;
    }
    AIDA::IFitResult* res = BatchLab_BaseFitter_fit(*this,fPrinter,
      const_cast<AIDA::ITuple&>(*tuple),*evaluator,*filter,aFunction);
    delete evaluator;
    delete filter;
    return res;
  }}

  Lib::Out out(fPrinter);
  out << "BatchLab::BaseFitter::fit :" 
      << " passed AIDA::IFitData connected to an unknown source."
      << Lib::endl;
  return 0;
}
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& aFitData
,const std::string& aModel
) 
{
  //FIXME : there is no IFitData::cast !
  //BatchLab::FitData* fitData = Lib_Cast(aFitData,BatchLab::FitData);
  BatchLab::FitData* fitData = dynamic_cast<BatchLab::FitData*>(&aFitData);
  if(!fitData) {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :" 
        << " passed AIDA::IFitData is not a BatchLab::FitData."
        << Lib::endl;
    return 0;
  }

 {const AIDA::IHistogram1D* histo = fitData->histogram1D();
  if(histo) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,
              fFunctionFactory,fCounter,
              *const_cast<AIDA::IHistogram1D*>(histo),aModel);
  }}

/* FIXME
 {const AIDA::IHistogram2D* histo = fitData->histogram2D();
  if(histo) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,
              fFunctionFactory,fCounter,
              *const_cast<AIDA::IHistogram2D*>(histo),aModel);
  }}

 {int x,v;
  const AIDA::IDataPointSet* dps = fitData->dps(x,v);
  if(dps) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,
              fFunctionFactory,fCounter,
              *const_cast<AIDA::IDataPointSet*>(dps),x,v,aModel);
  }}
*/

  std::vector<std::string> cols;
  const AIDA::ITuple* tuple = fitData->tuple(cols);
  if(tuple) {
    if(cols.size()!=1) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :" 
          << " can't handle today more than one column."
          << Lib::endl;
      return 0;
    }
    AIDA::IEvaluator* evaluator = fTupleFactory.createEvaluator(cols[0]);
    if(!evaluator) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :" 
          << " can't create evaluator."
          << Lib::endl;
      return 0;
    }
    AIDA::IFilter* filter = fTupleFactory.createFilter("");
    if(!filter) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :" 
          << " can't create filter."
          << Lib::endl;
      return 0;
    }
    AIDA::IFitResult* res = BatchLab_BaseFitter_fit(*this,fPrinter,
                  fFunctionFactory,fCounter,
                  const_cast<AIDA::ITuple&>(*tuple),*evaluator,*filter,aModel);
    delete evaluator;
    delete filter;
    return res;
  }
  Lib::Out out(fPrinter);
  out << "BatchLab::BaseFitter::fit :" 
      << " passed AIDA::IFitData connected to an unknown source."
      << Lib::endl;
  return 0;
}
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& /*d*/
,const std::string& /*model*/
,const std::vector<double>& /*initialParameters*/
) 
{
  return 0;
}
#else
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& //fitData
,AIDA::IFunction& //func
,const std::string& //range
) {
  return 0;
}
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& //fitData
,const std::string& //model
,const std::string& //range
) {
  return 0;
}
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& //fitData
,const std::string& //model
,const std::vector<double>& //initialParams
,const std::string& //range
) {
  return 0;
}

#endif
