#ifndef Lib_Functions_h
#define Lib_Functions_h

// Inheritance :
#include <Slash/Data/IFunction.h>

#include <complex>
#include <vector>
#include <string>

namespace Slash {namespace Core {class IWriter;}}

#include <Lib/Processor.h>

namespace Lib {

namespace Function {

class Gaussian {
public:
  Gaussian(double = 1,double = 0,double = 1);
  Gaussian(const Gaussian&);
  Gaussian& operator=(const Gaussian&);
  double value(double) const;
  void set(double,double,double);
  void setScale(double);
  void setMean(double);
  void setSigma(double);
private:
  double fScale,fMean,fSigma;
};

class BreitWigner {
public:
  BreitWigner(double = 1,double = 0,double = 1);
  BreitWigner(const BreitWigner&);
  BreitWigner& operator=(const BreitWigner&);
  double value(double) const;
  void set(double,double,double);
  void setHeight(double);
  void setCenter(double);
  void setWidth(double);
private:
  double fHeight,fCenter,fWidth;
};

class Exponential {
public:
  Exponential(double = 0,double = 1);
  Exponential(const Exponential&);
  Exponential& operator=(const Exponential&);
  double value(double) const;
  void set(double,double);
  void setA(double);
  void setB(double);
private:
  double fA,fB;
};

class Polynomial {
public:
  Polynomial(const std::vector<double>&);
  Polynomial(const Polynomial&);
  Polynomial& operator=(const Polynomial&);
  double value(double) const;
  bool setParameters(const std::vector<double>&);
  void setParameter(unsigned int,double);
private:
  std::vector<double> fPs;
};

class Cauchy {
public:
  Cauchy(double = 1,double = 0,double = 1);
  Cauchy(const Cauchy&);
  Cauchy& operator=(const Cauchy&);
  double value(double) const;
  void set(double,double,double);
  void setHeight(double);
  void setCenter(double);
  void setWidth(double);
private:
  double fHeight,fCenter,fWidth;
};

class Scripted {
public:
  Scripted(const std::string&,Slash::Core::IWriter&);
  Scripted(const Scripted&);
  virtual ~Scripted();
  Scripted& operator=(const Scripted&);
  //
  bool isValid() const;
  unsigned int dimension() const;
  std::string script() const;
  std::vector<std::string> parameterNames() const;
  std::vector<double> parameterValues() const;
  std::vector<std::string> variableNames() const;
  //
  bool set(const std::string&);
  bool setParameters(const std::vector<double>&);
  bool value(const std::vector<double>&,double&,std::string&) const;
private:
  static bool isVariable(const std::string&,int&);
  void reset();
private:
  std::string fScript;
  bool fIsValid;
  unsigned int fDimension;
  std::vector<std::string> fParameterNames;
  std::vector<double> fParameterValues;
  std::vector<std::string> fVariableNames;
  Slash::Core::IWriter& fPrinter;
  Processor fProcessor;
  std::vector<Lib::Value> fInput;
};

class Composite {
public:
  Composite(const std::string&,Slash::Core::IWriter&);
  Composite(const Composite&);
  virtual ~Composite();
  Composite& operator=(const Composite&);
  //
  bool isValid() const;
  unsigned int dimension() const;
  std::string script() const;
  std::vector<std::string> parameterNames() const;
  std::vector<double> parameterValues() const;
  std::vector<std::string> variableNames() const;
  //
  bool set(const std::string&);
  bool setParameters(const std::vector<double>&);
  bool value(const std::vector<double>&,double&,std::string&) const;
private:
  static bool isVariable(const std::string&,int&);
  void reset();
private:
  std::string fScript;
  Scripted fScripted;
  Slash::Core::IWriter& fPrinter;
};

class Compiled : public virtual Slash::Data::IFunction {
public: // IFunction :
  virtual std::string file() const;
  virtual std::string model() const;
  virtual void* address() const;
  virtual unsigned int dimension() const;
  virtual unsigned int numberOfParameters() const;
  virtual bool value(const std::vector<double>&,double&) const;
  virtual bool value(const std::vector<Slash::Core::IValue*>&,
                     Slash::Core::IValue&) const;
  virtual std::vector<Slash::Core::IValue*> signature(Slash::Core::IValue*&) const;
public:
  Compiled(Slash::Core::IWriter&,
           const std::string&,
           const std::string&,
           void*,
           const std::vector<Lib::Value>& args,
           unsigned int);
  virtual ~Compiled();
private:
  Slash::Core::IWriter& fPrinter;
  std::string fFile;
  std::string fModel;
  void* fAddress;
  std::vector<unsigned int> fDims;
  unsigned int fParn;
  std::vector<Slash::Core::IValue*> fValues;
  Slash::Core::IValue* fReturn;
};

class InsidePolygon {
public:
  typedef std::pair<double,double> Point;
public:
  InsidePolygon();
  //InsidePolygon(const std::vector<Point>&);
  InsidePolygon(const InsidePolygon&);
  InsidePolygon& operator=(const InsidePolygon&);
public:
  double value(const Point&) const;
  bool set(const std::vector<Point>&);
  bool set(unsigned int,const Point&);
  const std::vector<Point>& points() const;
private:
  inline static double is_left(const Point& P0,const Point& P1,const Point& P2){
    return ( (P1.first - P0.first) * (P2.second - P0.second)
           - (P2.first - P0.first) * (P1.second - P0.second) );
  }
  static bool inside(const Point&,const std::vector<Point>&);
private:
  std::vector<Point> fPoints;
};

class InsideEllipse {
public:
  InsideEllipse();
  InsideEllipse(const InsideEllipse&);
  InsideEllipse& operator=(const InsideEllipse&);
public:
  bool set_from_a_b(double,double);
  double value(double,double) const;
  double a() const;
  double b() const;
private:
  double f_a;
  double f_b;
};

} //Function

} //Lib

#endif




