// this :
#include <Lib/DLL.h>

#include <Lib/Debug.h>

#include <Slash/Tools/smanip.h>

#include <stdlib.h>

#ifdef WIN32
#include <windows.h>
#endif

#if defined(__alpha) || defined(__linux) || defined(__sun) || (defined(__CYGWIN__) && defined(__GNUC__))
#define DLOPEN
#define DLCLOSE //try that by default.
#endif

#if defined(__APPLE__)
//#define NSOPEN
//Since Tiger (10.4), dlopen comes with the system.
#define DLOPEN
#define DLCLOSE
#endif

#if defined(__APPLE__)
#include <vector>
static std::vector<std::string> paths() {
  const char* env = ::getenv("DYLD_LIBRARY_PATH");
  if(!env) env = ::getenv("GDB_DYLD_LIBRARY_PATH");
  std::string senv = (env==NULL ? std::string("") : std::string(env));
  return Slash::words(senv,":");
}
#endif

#ifdef DLOPEN
#include <dlfcn.h>
#endif

#ifdef NSOPEN
static void* nsopen(const std::string&,std::string&);
static void* nssym(void*,const std::string&);
static bool nsclose(void*);
#endif

Lib::DLL::DLL(
 const std::string& aName
,std::ostream& aOut
,int aVerboseLevel
,bool aQuiet
)
:fModule(0)
,fOut(aOut)
,fName(aName)
,fVerboseLevel(aVerboseLevel)
{
  // aName is the full name (path + libname).
  Debug::increment("Lib::DLL");
  if(!open()) {
    if(!aQuiet) {
      fOut << "Lib::DLL::DLL :" 
           << " can't load " << fName << " library." << std::endl;
      fOut << "Lib::DLL::DLL :" 
           << " error  : " << fError 
           << std::endl;
    }
  }
}

Lib::DLL::~DLL(){
  if (fModule) {
    if(fVerboseLevel) {
      fOut << "Lib::DLL::~DLL :" 
           << " close " << Slash::sout(fName) << "..." << std::endl;
    }
#ifdef DLOPEN
#ifdef DLCLOSE
    if(::dlclose(fModule)) {
      fOut << "Lib::DLL::~DLL :" 
           << " error  : " << ::dlerror() << std::endl;
    }
#endif
#endif
#ifdef NSOPEN
    nsclose(fModule);
#endif
#ifdef WIN32
    ::FreeLibrary((HMODULE)fModule);
#endif
    if(fVerboseLevel) {
      fOut << "Lib::DLL::~DLL :" 
           << " close " << Slash::sout(fName) << " done." << std::endl;
    }
  }
  fModule = 0;
  Debug::decrement("Lib::DLL");
}

bool Lib::DLL::open(){
  fError = "";
  std::string dll = fName;  
  if(fVerboseLevel) {
    fOut << "Lib::DLL::DLL :" 
         << " open " 
         << Slash::sout(dll) << "..." << std::endl;
  }
#ifdef DLOPEN
 {std::string path,name,suffix;
  Slash::path_name_suffix(fName,path,name,suffix);
  if(path.empty()) { //Relative path
    std::string suffix = Slash::suffix(dll);
    if(suffix.empty()) {
      // Only the name "Xxx" had been given, we have to
#if defined(__CYGWIN__) && defined(__GNUC__)
      // build the name "Xxx.dll" :      
      dll = fName + ".dll";
#elif defined(__APPLE__)
      dll = fName + ".bundle";
#else
      // build the name "libXxx.so" :      
      dll = "lib" + fName + ".so";
#endif
    }
  }}
  int flags = RTLD_NOW | RTLD_GLOBAL;
  fModule = ::dlopen(dll.c_str(),flags);
#if defined(__APPLE__)
  if(!fModule) {
    std::vector<std::string> ps = paths(); 
    unsigned int pathn = ps.size();
    for(unsigned int index=0;index<pathn;index++) {
      dll = ps[index]+"/" + fName + ".bundle";
      if(fVerboseLevel) {
        fOut << "Lib::DLL::DLL :" 
             << " open " 
             << Slash::sout(dll) << "..." << std::endl;
      }
      fModule = ::dlopen(dll.c_str(),flags);
      if(fModule) break;
    }
  }
#endif
  const char* cerror;
  if ((cerror=dlerror())!=NULL) fError = cerror;
#endif
#ifdef NSOPEN
  fModule = nsopen(dll,fError);
  if(!fModule) { //try <fName>.bundle
    std::vector<std::string> paths = paths(); 
    unsigned int pathn = paths.size();
    for(unsigned int index=0;index<pathn;index++) {
      dll = paths[index]+"/" + fName + ".bundle";
      if(fVerboseLevel) {
        fOut << "Lib::DLL::DLL :" 
             << " open " 
             << Slash::sout(dll) << "..." << std::endl;
      }
      fModule = nsopen(dll,fError);
      if(fModule) break;
    }
  }
  if(!fModule) { //try lib<fName>.dylib
    std::vector<std::string> paths = paths(); 
    unsigned int pathn = paths.size();
    for(unsigned int index=0;index<pathn;index++) {
      dll = paths[index]+"/lib" + fName + ".dylib";
      if(fVerboseLevel) {
        fOut << "Lib::DLL::DLL :" 
             << " open " 
             << Slash::sout(dll) << "..." << std::endl;
      }
      fModule = nsopen(dll,fError);
      if(fModule) break;
    }
  }
  if(!fModule) { //try lib<fName>.so (used in Panoramix)
    std::vector<std::string> paths = paths(); 
    unsigned int pathn = paths.size();
    for(unsigned int index=0;index<pathn;index++) {
      dll = paths[index]+"/lib" + fName + ".so";
      if(fVerboseLevel) {
        fOut << "Lib::DLL::DLL :" 
             << " open " 
             << Slash::sout(dll) << "..." << std::endl;
      }
      fModule = nsopen(dll,fError);
      if(fModule) break;
    }
  }
#endif
#ifdef WIN32
  fModule = ::LoadLibrary(dll.c_str());
#endif
  if(!fModule) return false;
  if(fVerboseLevel) {
    fOut << "Lib::DLL::DLL :" 
         << " " << Slash::sout(dll) << " opened." << std::endl;
  }
  fDLL = dll;  
  return true;
}

bool Lib::DLL::isOpened() const {return fModule?true:false;}

const std::string& Lib::DLL::fileName() const {return fName;}

Lib::Procedure Lib::DLL::findProcedure(const std::string& aName,bool aQuiet) const {
  if(!fModule) return 0;
  std::string name = aName;
  Procedure p = 0;
  if(fVerboseLevel) {
    fOut << "Lib::DLL::findProcedure :" 
         << " search symbol " 
         << Slash::sout(name) << "..." << std::endl;
  }
#ifdef DLOPEN
  p = (Procedure)::dlsym(fModule,(char*)name.c_str());
  const char* cerror;
  if ((cerror=dlerror())!=NULL) {
    std::string serror = cerror;
    name = "_" + aName;
    p = (Procedure)::dlsym(fModule,(char*)name.c_str());
    if ((cerror=dlerror())!=NULL) {
      if(!aQuiet) {
        fOut << "Lib::DLL::findProcedure :" 
             << " problem when searching " << Slash::sout(aName)
             << " in " << Slash::sout(fDLL) << "."
             << " dlerror  : " << serror << std::endl;
      } 
    } 
  }
#endif
#ifdef WIN32
  p = (Procedure)::GetProcAddress((HMODULE)fModule,name.c_str());
  if (!p) {
    name = "_" + aName;
    p = (Procedure)::GetProcAddress((HMODULE)fModule,name.c_str());
    if (!p) {
      if(!aQuiet) {
        fOut << "Lib : Can't find symbol " << Slash::sout(aName) << "." 
             << std::endl;
      }
    }
  }
#endif
#ifdef NSOPEN
  p = (Procedure)nssym(fModule,name);
  if(!p) {
    name = "_" + aName;
    if(fVerboseLevel) {
      fOut << "Lib::DLL::findProcedure :" 
           << " search symbol" 
           << Slash::sout(name) << "..." << std::endl;
    }
    p = (Procedure)nssym(fModule,name);
    if (!p) {
      if(!aQuiet) {
        fOut << "Lib::DLL::findProcedure :" 
             << " problem when searching " << Slash::sout(aName)
             << " in " << Slash::sout(fDLL) << "."
             << std::endl;
      } 
    } 
  }
#endif
  if(p && fVerboseLevel) {
    fOut << "Lib::DLL::findProcedure :" 
         << " symbol " << Slash::sout(name) << " found." << std::endl;
  }
  return p;
}

void Lib::DLL::execute(const std::string& aName) {
  typedef void(*F)();
  F func = (F)findProcedure(aName);
  if(!func) {
    fOut << "Function " << aName << " not found" << std::endl;
    return;
  }
  func();
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool Lib::DLL::isValid(const std::string& aName,std::ostream& aOut){
  Lib::DLL* dll = new Lib::DLL(aName,aOut,0,true);
  bool status = dll->isOpened();
  delete dll;
  return status;
}


#ifdef NSOPEN
#include <mach-o/dyld.h>
void* nsopen(const std::string& aDLL,std::string& aError) {
  //Foro infos : Darwin> man NSModule, NSObjectFileImage
  NSObjectFileImage ofi = 0;
  NSObjectFileImageReturnCode code = 
    NSCreateObjectFileImageFromFile(aDLL.c_str(),&ofi);
  if(code!=NSObjectFileImageSuccess) {
    switch(code)        {
    case NSObjectFileImageSuccess:
      break;
    case NSObjectFileImageInappropriateFile:
      aError = "inappropriate file.";
      break;
    case NSObjectFileImageFailure:
      aError = "object file setup failure.";
      break;
    case NSObjectFileImageArch:
      aError =  "no object for this architecture.";
      break;
    case NSObjectFileImageFormat:
      aError = "bad object file format.";
      break;
    case NSObjectFileImageAccess:
      aError = "can't read object file.";
      break;
    default:
      aError = "unknown error from NSCreateObjectFileImageFromFile().";
      break;
    }
    return 0;
  }
  aError = "";
  NSModule module = NSLinkModule(ofi,aDLL.c_str(),FALSE);
  NSDestroyObjectFileImage(ofi);
  return module;
}

void* nssym(void* aModule,const std::string& aName) {
  NSModule module = (NSModule)aModule;
  NSSymbol symbol = NSLookupSymbolInModule(module,aName.c_str());
  if(symbol==NULL) return 0;
  return NSAddressOfSymbol(symbol);
}

bool nsclose(void* aModule) {
  NSModule module = (NSModule)aModule;
  NSUnLinkModule(module,FALSE);
  return true;
}
#endif
