// this :
#include <Lib/Session.h>

#include <Slash/Core/IManager.h>
#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>
#include <Lib/Debug.h>

Lib::Session::Session(Slash::Core::IWriter& aPrinter,bool aBalance,bool aDebugCheckClass)
:fPrinter(aPrinter)
,fStreamBuffer(aPrinter)
,f_cout(&fStreamBuffer)
,fBalance(aBalance)
,fVerboseLevel(0)
,fParameters("","")
{
  Lib::Debug::checkByClass(aDebugCheckClass);
  Lib::Debug::increment("Lib::Session");
}

Lib::Session::~Session(){
  rm_managers();
  Lib::Debug::decrement("Lib::Session");
  if(fBalance) Lib::Debug::balance(f_cout);
}

//private:
Lib::Session::Session(const Session& aFrom)
:fPrinter(aFrom.fPrinter)
,fStreamBuffer(aFrom.fPrinter)
,f_cout(&fStreamBuffer)
,fBalance(aFrom.fBalance)
,fParameters("","")
{}
//private:
Lib::Session& Lib::Session::operator=(const Session&){return *this;}

void* Lib::Session::cast(const std::string& aClass) const {
  // inheritance :
  if_Lib_SCast(Slash::Core::ISession)
  // this :
  else if_Lib_SCast(Lib::Session)
  else return 0;
}

bool Lib::Session::rm_managers(){
  std::vector<Slash::Core::IManager*>::iterator it;
  while(!fManagers.empty()) {
    it = fManagers.end();
    it--;
    if(fVerboseLevel) {
      Lib::Out out(fPrinter);
      out << "Lib::Session::~Session :"
          << " delete manager " << Lib::sout((*it)->name()) 
          << "." << Lib::endl;
    }
    Slash::Core::IManager* manager = *it;
    fManagers.erase(it); //WARNING : must be done before the below.
    delete manager; //WARNING : some ~manager may rm themself from the list.
  }
  return true;
}

void Lib::Session::setVerboseLevel(int aVerboseLevel){fVerboseLevel = aVerboseLevel;}
int Lib::Session::verboseLevel() const{return fVerboseLevel;}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void Lib::Session::addManager(Slash::Core::IManager* aManager){
  if(findManager(aManager->name())) {
    Lib::Out out(fPrinter);
    out << "Lib::Session::addManager :"
        << " WARNING : a manager with name "
        << Lib::sout(aManager->name()) << " already exists." 
        << Lib::endl;
  }
  if(fVerboseLevel) {
    Lib::Out out(fPrinter);
    out << "Lib::Session::addManager :"
        << " " << Lib::sout(aManager->name()) << "." 
        << Lib::endl;
  }
  fManagers.push_back(aManager);
}

void Lib::Session::removeManager(Slash::Core::IManager* aManager){
  if(fVerboseLevel) {
    Lib::Out out(fPrinter);
    out << "Lib::Session::removeManager :"
        << " " << Lib::sout(aManager->name()) << "." 
        << Lib::endl;
  }
  std::vector<Slash::Core::IManager*>::iterator it;
  for(it=fManagers.begin();it!=fManagers.end();) {
    if((*it)==aManager) {
      it = fManagers.erase(it);
    } else {
      ++it;
    }
  }
}

Slash::Core::IManager* Lib::Session::findManager(const std::string& aString,bool aForward) const {
  if(aForward) {
    std::vector<Slash::Core::IManager*>::const_iterator it;
    for(it=fManagers.begin();it!=fManagers.end();++it) {
      if(aString==(*it)->name()) return *it;
    }
  } else {
    std::vector<Slash::Core::IManager*>::const_reverse_iterator it;
    for(it=fManagers.rbegin();it!=fManagers.rend();++it) {
      if(aString==(*it)->name()) return *it;
    }
  }
  return 0;
}

void Lib::Session::destroyManager(const std::string& aName){
  std::vector<Slash::Core::IManager*>::iterator it;
  for(it=fManagers.begin();it!=fManagers.end();) {
    Slash::Core::IManager* manager = *it;
    if(manager->name()==aName) {
      it = fManagers.erase(it);
      if(fVerboseLevel) {
        Lib::Out out(fPrinter);
        out << "Lib::Session::destroyManager :"
            << " " << Lib::sout(aName) << "." 
            << Lib::endl;
      }
      delete manager;
    } else {
      ++it;
    }
  }
}

unsigned int Lib::Session::numberOfManagers() const{return fManagers.size();}

const Slash::Core::IManager* Lib::Session::manager(unsigned int aIndex) const{return fManagers[aIndex];}

std::vector<std::string> Lib::Session::managerNames() const{
  std::vector<std::string> names;
  std::vector<Slash::Core::IManager*>::const_iterator it;
  for(it=fManagers.begin();it!=fManagers.end();++it) {
    names.push_back((*it)->name());
  }
  return names;
}

std::vector< std::pair<std::string,std::string> > Lib::Session::arguments() const{return fArguments;}

void Lib::Session::setArguments(const std::vector< std::pair<std::string,std::string> >& aArgs){fArguments = aArgs;}

Slash::Core::IWriter& Lib::Session::printer(){return fPrinter;}
std::ostream& Lib::Session::cout(){return f_cout;}

bool Lib::Session::setParameter(const std::string& aKey,const std::string& aValue){
  if(aKey.empty()) return false;
  if(fParameters.isKey(aKey)) { 
    fParameters.setValue(aKey,aValue);
  } else {
    fParameters.add(0,aKey,"",aValue,"","");
  }
  return true;
}

bool Lib::Session::parameterValue(const std::string& aKey,std::string& aValue){
  if(fParameters.isKey(aKey)) {
    aValue = fParameters.value(aKey);
    return true;
  } else {
    aValue = "";
    return false;
  }
}

bool Lib::Session::removeParameter(const std::string& aKey){
  fParameters.remove(aKey);
  return true;
}

std::vector<std::string> Lib::Session::availableParameters() const{return fParameters.keys();}

// no implementation :
bool Lib::Session::load(const std::string&,const std::string&){return false;}
void Lib::Session::redirectOutput(bool){}
bool Lib::Session::flush(){return false;}

