
#include <Lib/SpaceAlgebra.h>

#include <math.h>

//////////////////////////////////////////////////////////////////////////////
// SpacePlane is all p such that fNormal . p - fDistance = 0
// FDistance from origin to plane: fDistance * fNormal is on the plane
//////////////////////////////////////////////////////////////////////////////
Lib::SpacePlane::SpacePlane(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
Lib::SpacePlane::SpacePlane(
 const A3d& aP0
,const A3d& aP1
,const A3d& aP2
)
//////////////////////////////////////////////////////////////////////////////
// Construct a plane given 3 points.
// Orientation is computed by taking (p1 - p0) x (p2 - p0) and
// pointing the normal in that direction.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  A3d P = aP1;
  P.subtract(aP0);
  A3d P2 = aP2;
  P2.subtract(aP0);
  fNormal = P.cross(P2);
  fNormal.normalize();
  fDistance = 
    fNormal.value0() * aP0.value0() + 
    fNormal.value1() * aP0.value1() + 
    fNormal.value2() * aP0.value2();
}
//////////////////////////////////////////////////////////////////////////////
Lib::SpacePlane::SpacePlane(
 const A3d& aNormal
,double aFDistance
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
// Construct a plane given normal and fDistance from origin along normal.
// Orientation is given by the normal vector n.
{
  fNormal = aNormal;
  fNormal.normalize();
  fDistance = aFDistance;
}
//////////////////////////////////////////////////////////////////////////////
Lib::SpacePlane::SpacePlane(
 const A3d& aNormal
,const A3d& aPoint
)
//////////////////////////////////////////////////////////////////////////////
// Construct a plane given normal and a point to pass through
// Orientation is given by the normal vector n.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fNormal = aNormal;
  fNormal.normalize();
  fDistance = 
    fNormal.value0() * aPoint.value0() + 
    fNormal.value1() * aPoint.value1() + 
    fNormal.value2() * aPoint.value2();
}
//////////////////////////////////////////////////////////////////////////////
void Lib::SpacePlane::offset(
 double aFDistance
)
//////////////////////////////////////////////////////////////////////////////
// Offset a plane by a given fDistance.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fDistance += aFDistance;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::SpacePlane::intersect(
 const SpaceLine& aSpaceLine
,A3d& aIntersection
) const
//////////////////////////////////////////////////////////////////////////////
// Intersect line and plane, returning true if there is an intersection
// false if line is parallel to plane
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const A3d& pos = aSpaceLine.position();
  const A3d& dir = aSpaceLine.direction();
  double d = fNormal.dot(dir);
  if(d==0.0F) return false;
  double t = (fDistance - fNormal.dot(pos))/d;
  aIntersection = dir;
  aIntersection.multiply(t);
  aIntersection.add(pos);
  //aIntersection = pos + t * dir;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::SpacePlane::isInHalfSpace(
 const A3d& aPoint
) const
//////////////////////////////////////////////////////////////////////////////
// Returns true if the given point is within the half-space
// defined by the plane
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //A3d pos = fNormal * fDistance;
  A3d pos = fNormal;
  pos.multiply(-fDistance);
  pos.add(aPoint);
  return (fNormal.dot(pos) >= 0.0F ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
const Lib::A3d& Lib::SpacePlane::normal(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  return fNormal; 
}
//////////////////////////////////////////////////////////////////////////////
double Lib::SpacePlane::distanceFromOrigin(
) const        
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  return fDistance; 
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Parametric description:
//  l(t) = pos + t * dir
//////////////////////////////////////////////////////////////////////////////
Lib::SpaceLine::SpaceLine(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
Lib::SpaceLine::SpaceLine(
 const Lib::A3d& aP0
,const Lib::A3d& aP1
)
//////////////////////////////////////////////////////////////////////////////
// Construct a line from two points lying on the line.  If you
// want to construct a line from a position and a direction, use
// SpaceLine(p, p + d).
// SpaceLine is directed from p0 to p1.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fPos = aP0;
  //fDir = aP1-aP0;
  fDir = aP0;
  fDir.multiply(-1);
  fDir.add(aP1);
  fDir.normalize();
}
//////////////////////////////////////////////////////////////////////////////
void Lib::SpaceLine::setValue(
 const Lib::A3d& aP0
,const Lib::A3d& aP1
)
//////////////////////////////////////////////////////////////////////////////
// Set that value!
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fPos = aP0;
  //fDir = aP1-aP0;
  fDir = aP0;
  fDir.multiply(-1);
  fDir.add(aP1);
  fDir.normalize();
}
//////////////////////////////////////////////////////////////////////////////
const Lib::A3d& Lib::SpaceLine::position(
) const        
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  return fPos; 
}
//////////////////////////////////////////////////////////////////////////////
const Lib::A3d& Lib::SpaceLine::direction(
) const        
//////////////////////////////////////////////////////////////////////////////
// Accessors
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  return fDir; 
}

/* trash :

A3d operator*(const A3d&,double);
A3d operator*(double,const A3d&);
A3d operator/(const A3d&,double);
A3d operator+(const A3d&,const A3d&);
A3d operator-(const A3d&,const A3d&);
int operator==(const A3d&,const A3d&);
int operator!=(const A3d&,const A3d&);

int operator==(const SpacePlane&,const SpacePlane&);
int operator!=(const SpacePlane&,const SpacePlane&);

//////////////////////////////////////////////////////////////////////////////
Lib::A3d operator *(
 const Lib::A3d& aVec
,double aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return Lib::A3d(aVec[0] * aValue,
                     aVec[1] * aValue,
                     aVec[2] * aValue);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A3d operator *(
 double aValue
,const Lib::A3d& aVec
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  return operator*(aVec,aValue);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A3d operator /(
 const Lib::A3d& aVec
,double aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return Lib::A3d(aVec[0] / aValue,
                     aVec[1] / aValue,
                     aVec[2] / aValue);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A3d operator +(
 const Lib::A3d& aV1
,const Lib::A3d& aV2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return Lib::A3d(aV1[0] + aV2[0],
                     aV1[1] + aV2[1],
                     aV1[2] + aV2[2]);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A3d operator -(
 const Lib::A3d& aV1
,const Lib::A3d& aV2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return Lib::A3d(aV1[0] - aV2[0],
                     aV1[1] - aV2[1],
                     aV1[2] - aV2[2]);
}
//////////////////////////////////////////////////////////////////////////////
int operator ==(
 const Lib::A3d& aV1
,const Lib::A3d& aV2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return ( (aV1[0] == aV2[0]) && 
           (aV1[1] == aV2[1]) &&
           (aV1[2] == aV2[2])) ;
}
//////////////////////////////////////////////////////////////////////////////
int operator !=(
 const Lib::A3d& aV1
,const Lib::A3d& aV2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  return !operator==(aV1,aV2);
}

//////////////////////////////////////////////////////////////////////////////
int operator ==(
 const Lib::SpacePlane& aP1
,const Lib::SpacePlane& aP2
)
//////////////////////////////////////////////////////////////////////////////
// Equality/inequality comparison operators
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const Lib::A3d& N1 = aP1.normal();
  const Lib::A3d& N2 = aP2.normal();
  if ( ! operator==(N1,N2) ) return 0;
  if(aP1.distanceFromOrigin()!=aP2.distanceFromOrigin()) return 0;
  return 1;
}
//////////////////////////////////////////////////////////////////////////////
int operator!=(
 const Lib::SpacePlane& aP1
,const Lib::SpacePlane& aP2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  return !operator==(aP1,aP2); 
}

//////////////////////////////////////////////////////////////////////////////
bool Lib::SpaceLine::closestPoints(
 const SpaceLine& aSpaceLine2
,Lib::A3d& aPtOnThis
,Lib::A3d& aPtOnSpaceLine2
) const
//////////////////////////////////////////////////////////////////////////////
// Find closest points between the two lines. Return false if they are 
// parallel, otherwise return true.
//  p1 + t1 * d1 
//  p2 + t2 * d2 
//  (p2 - p1 + t2 * d2 - t1 * d1) * d1 = 0
//  (p2 - p1 + t2 * d2 - t1 * d1) * d2 = 0
//  t2 * (d2.d1) - t1 = (p1 - p2).d1
//  t2 - t1 * (d1.d2) = (p1 - p2).d2
//  
//  delta = 1 - (d1.d2)**2 
//
//  t2 = [ d2.(p1-p2) - d1.(p1-p2) * (d1.d2)]/delta
//  t1 = [-d1.(p1-p2) + d2.(p1-p2) * (d1.d2)]/delta
//
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const A3d& pos2 = aSpaceLine2.position();
  const A3d& dir2 = aSpaceLine2.direction();

  double dd = fDir.dot(dir2);
  double delta = 1. - dd * dd;

  if(delta==0.0F) return false;

  double t2 = (  dir2.dot(fPos - pos2) - fDir.dot(fPos - pos2) * dd)/delta;
  double t  = ( -fDir.dot(fPos - pos2) + dir2.dot(fPos - pos2) * dd)/delta;

  aPtOnThis  = fPos + t  * fDir;
  aPtOnSpaceLine2 = pos2 + t2 * dir2;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
Lib::A3d Lib::SpaceLine::closestPoint(
 const Lib::A3d& aPoint
) const
//////////////////////////////////////////////////////////////////////////////
// Returns the closest point on the line to the given point.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double t = (aPoint - fPos).dot(fDir);
  return (fPos + t * fDir); 
}

public:
  bool closestPoints(const SpaceLine&,A3d&,A3d&) const;
  A3d closestPoint(const A3d&) const;
  bool intersect(double,const A3d&) const;
  bool intersect(double,const A3d&,const A3d&,A3d&) const;
  bool intersect(const A3d&,const A3d&,const A3d&,
                 A3d&,A3d&,bool&) const;

//////////////////////////////////////////////////////////////////////////////
bool Lib::SpaceLine::intersect( 
 double aRadius //aAngle ? Used by SoRayPickAction
,const Lib::A3d& aPoint 
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  A3d point = closestPoint(aPoint);
  double length = (aPoint-point).length();
  if(length>aRadius) return false;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::SpaceLine::intersect( 
 double aRadius //aAngle ? Used by SoRayPickAction
,const Lib::A3d& aP0
,const Lib::A3d& aP1
,A3d& aPoint
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SpaceLine line(aP0,aP1);
  A3d pThis;
  if(closestPoints(line,pThis,aPoint)==false) return false;
  double length = (aPoint-pThis).length();
  if(length>aRadius) return false;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::SpaceLine::intersect( 
 const Lib::A3d& aV0
,const Lib::A3d& aV1
,const Lib::A3d& aV2
,Lib::A3d& aPoint
,Lib::A3d& aBarycentric
,bool& aFront 
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aBarycentric = (aV0 + aV1 + aV2)/3.;
  A3d v01 = aV1 - aV0;
  A3d v12 = aV2 - aV1;
  A3d v20 = aV0 - aV2;
  SpacePlane plane(aV0,aV1,aV2);
  const A3d& normal = plane.normal();
  aFront = normal.dot(A3d(0.,0.,1.))>=0 ? true : false;
  if(plane.intersect(*this,aPoint)==false) return false;
  // aPoint inside triangle ? 
  if(normal.dot(v01.cross(aPoint-aV0))<0) return false;
  if(normal.dot(v12.cross(aPoint-aV1))<0) return false;
  if(normal.dot(v20.cross(aPoint-aV2))<0) return false;
  return true;
}

*/
